#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "api.h"
#include "poly.h"
#include "test_dist.h"
#include "benchmarks.h"
#include "cpucycles.h"

#include "precomp.h"
#include "normaldist.h"
#include "samplerZ.h"
#include "randombytes.h"

#define PRINT_SIZE 8
#define REPEAT 10

#define PRINT_START_END(str, x)                                            \
  do                                                                       \
  {                                                                        \
    printf("* %s: ", str);                                                 \
    for (int i = 0; i < PRINT_SIZE; ++i)                                   \
      printf("%f%+fj  ", (x).coeffs[i].v, (x).coeffs[i + SOLMAE_D / 2].v); \
    printf("... ");                                                        \
    for (int i = SOLMAE_D / 2 - PRINT_SIZE; i < SOLMAE_D / 2; ++i)         \
      printf("%f%+fj  ", (x).coeffs[i].v, (x).coeffs[i + SOLMAE_D / 2].v); \
    printf("\n");                                                          \
  } while (0)

static void print_sk(secret_key *sk)
{
  PRINT_START_END("b10", sk->b10);
  PRINT_START_END("b11", sk->b11);
  PRINT_START_END("b20", sk->b20);
  PRINT_START_END("b21", sk->b21);
#if 0
  PRINT_START_END("GSO_b10", sk->GSO_b10);
  PRINT_START_END("GSO_b11", sk->GSO_b11);
  PRINT_START_END("GSO_b20", sk->GSO_b20);
  PRINT_START_END("GSO_b21", sk->GSO_b21);
  PRINT_START_END("sigma1", sk->sigma1);
  PRINT_START_END("sigma2", sk->sigma2);
  PRINT_START_END("beta10", sk->beta10);
  PRINT_START_END("beta11", sk->beta11);
  PRINT_START_END("beta20", sk->beta20);
  PRINT_START_END("beta21", sk->beta21);
#endif
}

void speed()
{
  secret_key sk;
  public_key pk;
  signature s;
  precomp(&sk);
  load_pk(&pk);

  uint8_t m[32] = {0x46, 0xb6, 0xc4, 0x83, 0x3f, 0x61, 0xfa, 0x3e, 0xaa, 0xe9, 0xad, 0x4a, 0x68, 0x8c, 0xd9, 0x6e, 0x22, 0x6d, 0x93, 0x3e, 0xde, 0xc4, 0x64, 0x9a, 0xb2, 0x18, 0x45, 0x2, 0xad, 0xf3, 0xc, 0x61};

  int iter = 10000;
  clock_t start_time = clock();
  uint64_t start = cpucycles();

  for (int i = 0; i < iter; ++i)
  {
    sign(m, &sk, &s);
  }
  uint64_t stop = cpucycles();
  clock_t stop_time = clock();

  double delta = (double)(stop - start) / iter;
  double delta_time = (double)(stop_time - start_time) / iter * 1.0e6 / CLOCKS_PER_SEC;
  printf("Average number of cycles per signature: %f (%f us)\n", delta, delta_time);
}

void speed_keygen()
{
  secret_key sk;
  public_key pk;

  int iter = 100;
  clock_t start_time = clock();
  uint64_t start = cpucycles();

  for (int i = 0; i < iter; ++i)
  {
    keygen_full(&sk, &pk);
  }
  uint64_t stop = cpucycles();
  clock_t stop_time = clock();

  double delta = (double)(stop - start) / iter;
  double delta_time = (double)(stop_time - start_time) / iter * 1.0e3 / CLOCKS_PER_SEC;
  printf("Average number of cycles per keygen: %f (%f ms)\n", delta, delta_time);
}

int _main()
{
  srand(time(0));
  seed_rng();
  printf("Hello world, signature is SOLMAE-%u\n", SOLMAE_D);
  secret_key sk;
  public_key pk;
  signature s;

  // uint8_t m[32] = {0x46, 0xb6, 0xc4, 0x83, 0x3f, 0x61, 0xfa, 0x3e, 0xaa, 0xe9, 0xad, 0x4a, 0x68, 0x8c, 0xd9, 0x6e, 0x22, 0x6d, 0x93, 0x3e, 0xde, 0xc4, 0x64, 0x9a, 0xb2, 0x18, 0x45, 0x2, 0xad, 0xf3, 0xc, 0x61};
  uint8_t m[32];

#if 0
  printf("Precomp.\n");
  precomp(&sk);

  load_pk(&pk);
  printf("Sign.\n");
  sign(m, &sk, &s);

  printf("Verify: %i\n", verify(m, &pk, &s));

  printf("Speed\n");
  speed();
  run_benchmarks();
  run_tests();
#endif

#if 0
  printf("\nTest keygen.\n");
  int trials = 0;
  for (int i = 0; i < 10; i++)
  {
    trials += keygen_full(&sk, &pk);
  }

  printf("[f,g] generated on average in %f trials with (alpha,alpha-eps)=(%f,%f)\n", ((double)trials) / 10., SOLMAE_ALPHA, SOLMAE_ALPHA - SOLMAE_ALPHAEPS);
  PRINT_START_END("f", sk.b10);
  PRINT_START_END("g", sk.b11);
#endif
  for (int n = 0; n < REPEAT; n++)
  {
    printf("[trial %d]\n", n + 1);

    randombytes(m, 32);
    keygen_full(&sk, &pk);

    print_sk(&sk);

    printf("message: ");
    for (int i = 0; i < 32; i++)
    {
      printf("%02x ", m[i]);
    }

    printf("\nSign.\n");
    sign(m, &sk, &s);

    printf("Verify: %i\n", verify(m, &pk, &s));
  }

#if 0
  printf("Speed keygen\n");
  speed_keygen();
#endif
#if 0
  FILE *precomp = fopen("precomp.data", "w");
  fprintf(precomp, "int8_t precomp_f[SOLMAE_D] = {");
  for(int i=0; i<SOLMAE_D; i++) {
      fprintf(precomp, "%d, ", sk.f[i]);
  } 
  fprintf(precomp, "};\n\n");
  fprintf(precomp, "int8_t precomp_g[SOLMAE_D] = {");
  for(int i=0; i<SOLMAE_D; i++) {
      fprintf(precomp, "%d, ", sk.g[i]);
  } 
  fprintf(precomp, "};\n\n");
  fprintf(precomp, "int8_t precomp_F[SOLMAE_D] = {");
  for(int i=0; i<SOLMAE_D; i++) {
      fprintf(precomp, "%d, ", sk.F[i]);
  } 
  fprintf(precomp, "};\n\n");
  fprintf(precomp, "int8_t precomp_G[SOLMAE_D] = {");
  for(int i=0; i<SOLMAE_D; i++) {
      fprintf(precomp, "%d, ", sk.G[i]);
  } 
  fprintf(precomp, "};\n\n");
  fprintf(precomp, "int16_t precomp_hint[SOLMAE_D] = {");
  for(int i=0; i<SOLMAE_D; i++) {
      fprintf(precomp, "%d, ", pk.hint[i]);
  } 
  fprintf(precomp, "};\n\n");

  fclose(precomp);
#endif

  return 0;
}
